#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>

#include <mad.h>

#include "ss.h"


double *left = NULL;
double *right = NULL;
unsigned long int n_samples = 0;
int sample_rate = 0;
int errors = 0;

struct buffer {
	unsigned char const *start;
	unsigned long length;
	char verbose, shown;
	unsigned long t_bitrate;
	int n;
};

static enum mad_flow input(void *data, struct mad_stream *stream)
{
	struct buffer *buffer = data;

	if (!buffer -> length)
		return MAD_FLOW_STOP;

	mad_stream_buffer(stream, buffer -> start, buffer -> length);

	buffer -> length = 0;

	return MAD_FLOW_CONTINUE;
}

static inline double scale(mad_fixed_t sample)
{
	return mad_f_todouble(sample);
}

static enum mad_flow output(void *data, struct mad_header const *header, struct mad_pcm *pcm)
{
	unsigned int nchannels, cur_nsamples;
	mad_fixed_t const *left_ch, *right_ch;
	int loop;

	/* pcm->samplerate contains the sampling frequency */

	nchannels = pcm->channels;
	cur_nsamples  = pcm->length;
	left_ch   = pcm->samples[0];
	right_ch  = pcm->samples[1];
	sample_rate = pcm->samplerate;

	left = (double *)realloc(left, sizeof(double) * (n_samples + cur_nsamples + DEFAULT_SAMPLERATE));
	right = (double *)realloc(right, sizeof(double) * (n_samples + cur_nsamples + DEFAULT_SAMPLERATE));

	for(loop=0; loop<cur_nsamples; loop++)
	{
		left[n_samples + loop] = scale(*left_ch++);

		if (nchannels == 2)
			right[n_samples + loop] = scale(*right_ch++);
		else
			right[n_samples + loop] = left[n_samples + loop];
	}

	n_samples += cur_nsamples;

	return MAD_FLOW_CONTINUE;
}

static enum mad_flow error(void *data, struct mad_stream *stream, struct mad_frame *frame)
{
	struct buffer *buffer = data;

	if (buffer -> verbose)
	{
		fprintf(stderr, "decoding error 0x%04x (%s) at byte offset %ld\n",
			stream->error, mad_stream_errorstr(stream),
			stream->this_frame - buffer->start);
	}
	else
	{
		errors++;
	}

	/* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

	return MAD_FLOW_CONTINUE;
}

static enum mad_flow header(void *data, struct mad_header const *mheader, struct mad_pcm *ppcm)
{
	struct buffer *buffer = data;

	if (buffer -> verbose && buffer -> shown == 0)
	{
		buffer -> shown = 1;

		printf("Mpeg layer: ");
		if (mheader -> layer == MAD_LAYER_I)
			printf("I\n");
		else if (mheader -> layer == MAD_LAYER_II)
			printf("II\n");
		else if (mheader -> layer == MAD_LAYER_III)
			printf("III\n");
		else
			printf("unknown (%d)\n", mheader -> layer);

		printf("Mode      : ");
		if (mheader -> mode == MAD_MODE_SINGLE_CHANNEL)
			printf("single channel\n");
		else if (mheader -> mode == MAD_MODE_DUAL_CHANNEL)
			printf("dual channel\n");
		else if (mheader -> mode == MAD_MODE_JOINT_STEREO)
			printf("joint (MS/intensity) stereo\n");
		else if (mheader -> mode == MAD_MODE_STEREO)
			printf("normal LR stereo\n");
		else
			printf("unknown (%d)\n", mheader -> mode);

		printf("Samplerate: %d\n", mheader -> samplerate);
	}

	buffer -> t_bitrate += mheader -> bitrate;
	buffer -> n++;

	return MAD_FLOW_CONTINUE;
}


static int decode(unsigned char const *start, unsigned long length, char verbose)
{
	struct buffer buffer;
	struct mad_decoder decoder;
	int result;

	if (verbose) printf("File size : %ldKB\n", length / 1024);

	/* initialize our private message structure */
	buffer.start   = start;
	buffer.length  = length;
	buffer.verbose = verbose;
	buffer.shown   = 0;
	buffer.t_bitrate = 0;
	buffer.n       = 0;

	errors = 0;

	/* configure input, output, and error functions */
	mad_decoder_init(&decoder, &buffer,
			input, header, 0 /* filter */, output,
			error, 0 /* message */);

	/* start decoding */
	result = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

	/* release the decoder */
	mad_decoder_finish(&decoder);

	if (verbose) printf("Average bitrate: %ld\n", buffer.t_bitrate / buffer.n);

	if (errors) printf("Number of decoding errors: %d\n", errors);

	return result;
}

void load_mp3(char *filename, char verbose, double **pleft, double **pright, int *psample_rate, unsigned long int *pn_samples)
{
	struct stat stat;
	void *data;
	int fd = open(filename, O_RDONLY);
	if (fd == -1)
	{
		fprintf(stderr, "Cannot open file %s: %s\n", filename, strerror(errno));
		exit(1);
	}

	if (fstat(fd, &stat) == -1)
	{
		fprintf(stderr, "Failed to fstat(): %s\n", strerror(errno));
		exit(1);
	}

	data = mmap(NULL, stat.st_size, PROT_READ, MAP_SHARED, fd, 0);
	if (data == MAP_FAILED)
	{
		fprintf(stderr, "Failed to mmap(): %s\n", strerror(errno));
		exit(1);
	}

	left = right = NULL;
	n_samples = 0;
	decode(data, stat.st_size, verbose);

	if (munmap(data, stat.st_size) == -1)
	{
		fprintf(stderr, "Failed to munmap(): %s\n", strerror(errno));
		exit(1);
	}

	close(fd);

	*pleft = left;
	*pright = right;
	*psample_rate = sample_rate;
	*pn_samples = n_samples;
}
