#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <sndfile.h>

#include <fftw3.h>
#include "fft.h"

#include "mp3.h"
#include "ss.h"

double pearson_correlation(double *d1, double *d2, int n)
{
	int loop;
	double avg_d1=0, avg_d2=0;
	double d1_sd_sq=0, d2_sd_sq=0, d1_d2_sd_sq=0;

	for (loop=0; loop<n; loop++)
	{
		avg_d1 += d1[loop];
		avg_d2 += d2[loop];
	}
	avg_d1 /= n;
	avg_d2 /= n;

	for (loop=0; loop<n; loop++)
	{
		double d1_sd = d1[loop]-avg_d1;
		double d2_sd = d2[loop]-avg_d2;
		d1_sd_sq += d1_sd*d1_sd;
		d2_sd_sq += d2_sd*d2_sd;
		d1_d2_sd_sq += d1_sd*d2_sd;
	}

	return d1_d2_sd_sq / sqrt(d1_sd_sq * d2_sd_sq);
}

void * mymalloc(int size)
{
	void * dummy = malloc(size);

	if (!dummy)
	{
		fprintf(stderr, "Failed to allocate %d bytes of memory!\n", size);
		exit(1);
	}

	return dummy;
}

void add_to_list(struct file **list, int *list_length, char *fname)
{
	*list = (struct file *)realloc(*list, (*list_length + 1) * sizeof(struct file));
	if (!*list)
	{
		fprintf(stderr, "Memory allocation error\n");
		exit(1);
	}

	(*list)[*list_length].filename = strdup(fname);
	if (!(*list)[*list_length].filename)
	{
		fprintf(stderr, "Memory allocation error\n");
		exit(1);
	}

	(*list_length)++;
}

void load_list_from_file(struct file **list, int *list_length, char *list_fname)
{
	FILE *fin = fopen(list_fname, "r");
	if (!fin)
	{
		fprintf(stderr, "Failed to open list-file %s: %s\n", list_fname, strerror(errno));
		exit(1);
	}

	while(!feof(fin))
	{
		char buffer[PATH_MAX];	/* replace by dynamic buffer sysconf etc. */
		char *lf;

		if (!fgets(buffer, PATH_MAX, fin))
			break;

		lf = strchr(buffer, '\n');
		if (lf) *lf = 0x00;

		add_to_list(list, list_length, buffer);
	}

	fclose(fin);
}

void load_file(char *filename, char verbose, double **left, double **right, int *sample_rate, unsigned long int *n_samples)
{
	char is_mp3 = 0;
	char *dot = strrchr(filename, '.');
	if (dot)
	{
		if (strcasecmp(dot, ".mp3") == 0)
			is_mp3 = 1;
	}

	if (is_mp3)
	{
		if (verbose) printf("%s is an MP3 file\n", filename);
		load_mp3(filename, verbose, left, right, sample_rate, n_samples);
	}
	else
	{
		SF_INFO file_in;
		SNDFILE *file_in_handle;
		unsigned long int loop, buffer_size;
		double *buffer_in;

		/* open file */
		file_in.format = 0;
		file_in_handle = sf_open(filename, SFM_READ, &file_in);
		if (!file_in_handle)
		{
			fprintf(stderr, "Failed to open file '%s': %s\n", filename, sf_strerror(file_in_handle));
			exit(1);
		}

		if (verbose)
		{
			int format = file_in.format & SF_FORMAT_TYPEMASK;
			int subformat = file_in.format & 0xff;
			int duration = file_in.frames / file_in.samplerate;

			printf("Duration  : %d:%02d\n", duration / 60, duration % 60);
			printf("Samplerate: %d\n", file_in.samplerate);
			printf("Channels  : %d\n", file_in.channels);
			printf("Format    : ");

			if (file_in.channels != 1 && file_in.channels != 2)
			{
				fprintf(stderr, "File must have either 1 or 2 channels!\n");
				exit(1);
			}

			if (subformat == SF_FORMAT_PCM_S8)
				printf("8 bit PCM ");
			else if (subformat == SF_FORMAT_PCM_16)
				printf("16 bit PCM ");
			else if (subformat == SF_FORMAT_PCM_24)
				printf("24 bit PCM ");
			else if (subformat == SF_FORMAT_PCM_32)
				printf("32 bit PCM ");
			else if (subformat >= 0x10 && subformat <= 0x43)
				printf("compressed audio ");
			else
				printf("unknown ");

			if (format == SF_FORMAT_WAV)
				printf("Microsoft WAV format (little endian).");
			else if (format == SF_FORMAT_AIFF)
				printf("Apple/SGI AIFF format (big endian).");
			else if (format == SF_FORMAT_AU)
				printf("Sun/NeXT AU format (big endian).");
			else if (format == SF_FORMAT_RAW)
				printf("RAW PCM data.");
			else if (format == SF_FORMAT_PAF)
				printf("Ensoniq PARIS file format.");
			else if (format == SF_FORMAT_SVX)
				printf("Amiga IFF / SVX8 / SV16 format.");
			else if (format == SF_FORMAT_NIST)
				printf("Sphere NIST format.");
			else if (format == SF_FORMAT_VOC)
				printf("VOC files.");
			else if (format == SF_FORMAT_IRCAM)
				printf("Berkeley/IRCAM/CARL");
			else if (format == SF_FORMAT_W64)
				printf("Sonic Foundry's 64 bit RIFF/WAV");
			else if (format == SF_FORMAT_MAT4)
				printf("Matlab (tm) V4.2 / GNU Octave 2.0");
			else if (format == SF_FORMAT_MAT5)
				printf("Matlab (tm) V5.0 / GNU Octave 2.1");
			else if (format == SF_FORMAT_PVF)
				printf("Portable Voice Format");
			else if (format == SF_FORMAT_XI)
				printf("Fasttracker 2 Extended Instrument");
			else if (format == SF_FORMAT_HTK)
				printf("HMM Tool Kit format");
			else if (format == SF_FORMAT_SDS)
				printf("Midi Sample Dump Standard");
			else if (format == SF_FORMAT_AVR)
				printf("Audio Visual Research");
			else if (format == SF_FORMAT_WAVEX)
				printf("MS WAVE with WAVEFORMATEX");
			else
				printf("unknown");
		}

		*sample_rate = file_in.samplerate;
		*n_samples = file_in.frames;

		buffer_size = sizeof(double) * (file_in.frames + file_in.samplerate);
		*left  = (double *)malloc(buffer_size);
		memset(*left, 0x00, buffer_size);
		*right = (double *)malloc(buffer_size);
		memset(*right, 0x00, buffer_size);
		buffer_in = (double *)malloc(sizeof(double) * 1024 * file_in.channels);

		for(loop=0; loop<file_in.frames; loop += 1024)
		{
			if (verbose) { printf("%f%%\r", ((double)loop * 100.0) / (double)file_in.frames); fflush(stdout); }

			/* read from file */
			if (sf_read_double(file_in_handle, buffer_in, 1024 * file_in.channels) == 0)
			{
				fprintf(stderr, "read error (%s) at %ld\n", sf_strerror(file_in_handle), loop);
				exit(1);
			}

			if (file_in.channels == 1)
			{
				int index;
				for(index=0; index<1024; index++)
					(*left)[loop + index] = (*right)[loop + index] = buffer_in[index];
			}
			else 
			{
				int index;
				for(index=0; index<1024; index++)
				{
					(*left)[loop + index]  = buffer_in[index * 2 + 0];
					(*right)[loop + index] = buffer_in[index * 2 + 1];
				}
			}
		}

		/* close file */
		sf_close(file_in_handle);

		free(buffer_in);
	}
}

void analyse_file(char *filename, char verbose, double **fft_out)
{
	int cnt = 0;
	int cur_sample;
	fft_info *fft_data;
	double *left_in = NULL, *right_in = NULL;
	unsigned long int n_samples, sample_index;
	int sample_rate;
	double *buffer_out = NULL;

	if (verbose) printf("Loading...\n");
	load_file(filename, verbose, &left_in, &right_in, &sample_rate, &n_samples);

	/* init fft library */
	fft_data = alloc_fft(sample_rate);

	/* allocate work buffer */
	buffer_out = (double *)mymalloc(sizeof(double) * sample_rate);
	*fft_out = (double *)mymalloc(sizeof(double) * max(sample_rate, DEFAULT_SAMPLERATE));
	memset(*fft_out, 0x00, sizeof(double) * max(sample_rate, DEFAULT_SAMPLERATE));

	if (verbose) printf("Analyzing...\n");
	for(sample_index = 0; sample_index < n_samples; sample_index += sample_rate)
	{
		if (verbose) { printf("%f%%\r", ((double)sample_index * 100.0) / (double)n_samples); fflush(stdout); }
		/* left */
		/* fft */
		do_fft(fft_data, &left_in[sample_index], sample_rate, buffer_out);
		/* add to fft_out[filenr] */
		for(cur_sample=0; cur_sample<sample_rate; cur_sample++)
			(*fft_out)[cur_sample] += buffer_out[cur_sample];
		/* right */
		/* fft */
		do_fft(fft_data, &right_in[sample_index], sample_rate, buffer_out);
		/* add to fft_out[filenr] */
		for(cur_sample=0; cur_sample<sample_rate; cur_sample++)
			(*fft_out)[cur_sample] += buffer_out[cur_sample];

		cnt+=2;
	}

	for(cur_sample=0; cur_sample<sample_rate; cur_sample++)
		(*fft_out)[cur_sample] /= (double)cnt;

	free_fft(fft_data);

	free(buffer_out);
	free(left_in);
	free(right_in);
}

void usage()
{
	fprintf(stderr, "-I file   load filenames from the file to the list\n");
	fprintf(stderr, "-i file   add 'file' to the list\n");
	fprintf(stderr, "-o file   file to write the results to\n");
	fprintf(stderr, "-O file   like -o only in playlist format\n");
	fprintf(stderr, "-f file   select 'file' to use as the first song, else\n");
	fprintf(stderr, "          a song is selected at random\n");
	fprintf(stderr, "-v        be verbose\n");
	fprintf(stderr, "-h        this help & exit\n");
}

int main(int argc, char *argv[])
{
	struct file *list = NULL;
	int n_files = 0;
	int argindex, filenr;
	FILE *file_out_handle = NULL;
	char *file_out = NULL;
	char *file_first = NULL;
	int first_index = -1;
	double *first_file_fft = NULL;
	double *pcc = NULL;
	int *indexes = NULL;
	int loop;
	char verbose = 0, playlist = 0;

	printf("soundsort " VERSION ", (C) 2005-2009 by folkert@vanheusden.com\n\n");

	for(argindex=1; argindex<argc; argindex++)
	{
		if (strcmp(argv[argindex], "-I") == 0)	/* -I: load filenames from a list */
		{
			load_list_from_file(&list, &n_files, argv[++argindex]);
		}
		else if (strcmp(argv[argindex], "-i") == 0)	/* -i: next parameter is a filename */
		{
			add_to_list(&list, &n_files, argv[++argindex]);
		}
		else if (strcmp(argv[argindex], "-o") == 0)	/* -o: file to write the sorted list to */
		{
			file_out = argv[++argindex];
		}
		else if (strcmp(argv[argindex], "-O") == 0)	/* -o: file to write the sorted list to */
		{
			file_out = argv[++argindex];
			playlist = 1;
		}
		else if (strcmp(argv[argindex], "-f") == 0) /* -f: file to put first in the list */
		{
			file_first = argv[++argindex];
		}
		else if (strcmp(argv[argindex], "-v") == 0) /* -v: be verbose */
		{
			verbose = 1;
		}
		else if (strcmp(argv[argindex], "-h") == 0)
		{
			usage();
			return 0;
		}
		else if (argv[argindex][0] == '-')
		{
			fprintf(stderr, "%s is an unknown switch\n", argv[argindex]);
			return 1;
		}
		else
		{
			add_to_list(&list, &n_files, argv[argindex]);
		}
	}

	if (n_files < 1)
	{
		fprintf(stderr, "No files selected to analyze.\n");
		usage();
		return 1;
	}

	printf("%d filenames loaded.\n", n_files);

	if (!file_first)
	{
		srand(time(NULL));

		file_first = list[rand() % n_files].filename;

		printf("Using '%s' as the first file.\n", file_first);
	}

	for(filenr=0; filenr<n_files; filenr++)
	{
		if (strcmp(list[filenr].filename, file_first) == 0)
		{
			first_index = filenr;
			break;
		}
	}

	if (first_index == -1)
	{
		fprintf(stderr, "File '%s' not found in list?\n", file_first);
		fprintf(stderr, "You not only need to set what file to put at the top\n");
		fprintf(stderr, "of the list, you also need to add it to the list of\n");
		fprintf(stderr, "files to analyze using -i/-I\n");
		return 1;
	}
	else
	{
		printf("You've chosen '%s' as the first song in the list.\n", list[first_index].filename);
	}

	/* verify that all can be opened */
	for(filenr=0; filenr<n_files; filenr++)
	{
		int fd = open(list[filenr].filename, O_RDONLY);
		if (fd == -1)
		{
			fprintf(stderr, "Failed to open file %s: %s\n", list[filenr].filename, strerror(errno));
			return 1;
		}
		close(fd);
	}

	/* init */
	pcc = (double *)mymalloc(sizeof(double) * n_files);
	indexes = (int *)mymalloc(sizeof(int) * n_files);

	/* analyze */
	printf("Analyzing file: %s\n", list[first_index].filename);
	analyse_file(list[first_index].filename, verbose, &first_file_fft);

	for(filenr=0; filenr<n_files; filenr++)
	{
		double *cur_file_fft = NULL;

		/* init */
		indexes[filenr] = filenr;

		/* already analysed the first_file */
		if (filenr == first_index)
		{
			pcc[filenr] = 1.0;
			continue;
		}

		printf("Analyzing file %d: %s\n", filenr + 1, list[filenr].filename);
		fflush(stdout);

		analyse_file(list[filenr].filename, verbose, &cur_file_fft);

		pcc[filenr] = pearson_correlation(cur_file_fft, first_file_fft, DEFAULT_SAMPLERATE / 2);

		free(cur_file_fft);
	}

	/* sort */
	printf("Sorting...\n");
	/* don't complain about this utterly stupid algorithm */
	for(loop=0; loop<(n_files - 1); loop++)
	{
		int loop2;

		printf("%3.2f%%\r", (filenr * 100.0) / (double)n_files);
		fflush(stdout);

		for(loop2=(loop + 1); loop2<n_files; loop2++)
		{
			if (pcc[loop2] > pcc[loop])
			{
				double dummy_pcc = pcc  [loop2];
				int dummy_index   = indexes[loop2];

				pcc  [loop2] = pcc  [loop];
				indexes[loop2] = indexes[loop];

				pcc  [loop] = dummy_pcc;
				indexes[loop] = dummy_index;
			}
		}
	}

	/* show */
	file_out_handle = stdout;
	if (file_out)
	{
		file_out_handle = fopen(file_out, "w");
		if (!file_out_handle)
			fprintf(stderr, "Error: could not create file '%s': %s (%d)\n", file_out, strerror(errno), errno);
	}
	for(filenr=0; filenr<n_files; filenr++)
	{
		if (playlist)
			fprintf(file_out_handle, "%s\n", list[indexes[filenr]].filename);
		else
			fprintf(file_out_handle, "%02d] %s %f\n", filenr + 1, list[indexes[filenr]].filename, pcc[filenr]);
	}
	if (file_out_handle != stdout)
		fclose(file_out_handle);

	return 0;
}
